#jinja2: variable_start_string: '[%', variable_end_string: '%]'
apiVersion: unfurl/v1alpha1
kind: Project
# this file is including into unfurl.yaml but is listed in .gitignore
# so you can add configuration that you don't want committed to the repository,
# such as secrets, local settings, and local instances.
contexts:
  defaults:
    # use "environment" to modify environment variables
    # keys are processed in order and a leader "+" and "-"
    # will copy or remove the variable from the environment
    # e.g.:
    # environment:
    #   -*:       # will remove all environment variables
    #   +HOME:    # will add HOME back
    #   FOO: bar  # set FOO
    environment: {}

    {%- if vaultpass %}
    secrets:
      attributes:
        # UNFURL_VAULT_DEFAULT_PASSWORD overrides this
        vault_default_password: "[% vaultpass %]"
    {% endif %}
{%if manifestPath %}
manifests:
  - file: [%manifestPath%]
    default: true
{% endif %}
