#jinja2: variable_start_string: '[%', variable_end_string: '%]'
apiVersion: unfurl/v1alpha1
kind: Home

# list of local repositories
locations: # like lock with localpath added
- path: # relative path in repo
 "name",
 "url",
 "revision",
 "initial"
 "origin"

# list of projects whose repositories are local to this machine
projects: # projects with remote origins are added to home, otherwise local/home
#   - url:
#     initialcommit:
