#jinja2: variable_start_string: '[%', variable_end_string: '%]'
apiVersion: unfurl/v1alpha1
kind: Project
+?include: local/unfurl.yaml
contexts:
  defaults:
    # this exposes this project's ensemble as "localhost" to all the projects
    # that inherit this context
    connections:
    # Below we define a few for the most common cloud providers
    # these connection types set their properties using
    # environments variables commonly associated with each cloud provider.
    # You can set or unset which environment variables are available
    # in the "environment" section in your project
    # or directly set the properties by adding them to your projects connections.
    # You can also delete the connection by setting its key to null (e.g. aws: )
    # or can add additional ones of the same type that can be explicitly referenced by name.
      aws:
        type: unfurl.relationships.ConnectsTo.AWSAccount

      gcp:
        type: unfurl.relationships.ConnectsTo.GCPProject

      azure:
        type: unfurl.relationships.ConnectsTo.Azure

      k8s:
        type: unfurl.relationships.ConnectsTo.K8sCluster

      # packet:
      #   type: unfurl.relationships.ConnectsTo.Packet
      #   properties:
      #     project: # required property, need to set this

    external:
      localhost:
        manifest:
          file: [% manifestPath | default("ensemble/ensemble.yaml", true) %]
        instance: localhost
