#jinja2: variable_start_string: '[%', variable_end_string: '%]'
apiVersion: unfurl/v1alpha1
kind: Home
+?include: local/home.yaml

# list of projects whose repository has a remote origin set
projects:
#   - url:
#     initialcommit:
